# -*- coding: utf-8 -*-
#   Copyright (C) 2008-2010, 2013, 2015, 2017-2018, 2020 Rocky Bernstein <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
""" Functions for working with Python frames"""

import inspect, linecache, os, re, xdis

import trepan.lib.bytecode as Mbytecode
import trepan.lib.printing as Mprint
import trepan.lib.format as Mformat
import trepan.lib.pp as Mpp

import os.path as osp
from trepan.lib.deparse import deparse_offset
from trepan.processor.cmdfns import deparse_fn

from xdis import PYTHON_VERSION, IS_PYPY, get_opcode

format_token = Mformat.format_token

_with_local_varname = re.compile(r"_\[[0-9+]\]")


def count_frames(frame, count_start=0):
    "Return a count of the number of frames"
    count = -count_start
    while frame:
        count += 1
        frame = frame.f_back
    return count


import reprlib as Mrepr

_re_pseudo_file = re.compile(r"^<.+>")


# Taken from 3.7 inspect.py. However instead of an object name, we
# start with the filename. Also, oddly getsourcefile wasn't
# stripping out __pycache__. Finally, we've adapted this
# so it works back to 3.0.
#
def getsourcefile(filename):
    """Return the filename that can be used to locate an object's source.
    Return None if no way can be identified to get the source.
    """
    all_bytecode_suffixes = (".pyc", ".pyo")
    if any(filename.endswith(s) for s in all_bytecode_suffixes):
        if osp.dirname(filename).endswith("__pycache__"):
            filename = osp.join(
                osp.dirname(osp.dirname(filename)), osp.basename(filename)
            )
        filename = osp.splitext(filename)[0] + ".py"
    elif any(filename.endswith(s) for s in [".abi3.so", ".so"]):
        return None
    if osp.exists(filename):
        return filename
    # only return a non-existent filename if the module has a PEP 302 loader
    if getattr(inspect.getmodule(object, filename), "__loader__", None) is not None:
        return filename
    # or it is in the linecache
    if filename in linecache.cache:
        return filename


def deparse_source_from_code(code):
    source_text = ""
    try:
        source_lines = deparse_fn(code).split("\n")
        source_text = ""
        for i, source_text in enumerate(source_lines):
            if len(source_text) > 0:
                break
        if len(source_lines) > 1:
            source_text += "..."
        source_text = '"%s"' % source_text
    except:
        pass
    return source_text


def format_stack_entry(
    dbg_obj, frame_lineno, lprefix=": ", include_location=True, color="plain"
):
    """Format and return a stack entry gdb-style.
    Note: lprefix is not used. It is kept for compatibility.
    """
    frame, lineno = frame_lineno
    filename = frame2file(dbg_obj.core, frame)

    s = ""
    if frame.f_code.co_name:
        funcname = frame.f_code.co_name
    else:
        funcname = "<lambda>"
        pass
    s = format_token(Mformat.Function, funcname, highlight=color)

    args, varargs, varkw, local_vars = inspect.getargvalues(frame)
    if "<module>" == funcname and ([], None, None,) == (
        args,
        varargs,
        varkw,
    ):
        is_module = True
        if is_exec_stmt(frame):
            fn_name = format_token(Mformat.Function, "exec", highlight=color)
            source_text = deparse_source_from_code(frame.f_code)
            s += " %s(%s)" % (
                format_token(Mformat.Function, fn_name, highlight=color),
                source_text,
            )
        else:
            fn_name = get_call_function_name(frame)
            if fn_name:
                source_text = deparse_source_from_code(frame.f_code)
                if fn_name:
                    s += " %s(%s)" % (
                        format_token(Mformat.Function, fn_name, highlight=color),
                        source_text,
                    )
            pass
    else:
        is_module = False
        try:
            parms = inspect.formatargvalues(args, varargs, varkw, local_vars)
        except:
            pass
        else:
            maxargstrsize = dbg_obj.settings["maxargstrsize"]
            if len(parms) >= maxargstrsize:
                parms = "%s...)" % parms[0:maxargstrsize]
                pass
            s += parms
        pass

    # Note: ddd can't handle wrapped stack entries (yet).
    # The 35 is hoaky though. FIXME.
    if len(s) >= 35:
        s += "\n    "

    if "__return__" in frame.f_locals:
        rv = frame.f_locals["__return__"]
        s += "->"
        s += format_token(Mformat.Return, Mrepr.repr(rv), highlight=color)
        pass

    if include_location:
        is_pseudo_file = _re_pseudo_file.match(filename)
        add_quotes_around_file = not is_pseudo_file
        if is_module:
            if filename == "<string>":
                s += " in exec"
            elif not is_exec_stmt(frame) and not is_pseudo_file:
                s += " file"
        elif s == "?()":
            if is_exec_stmt(frame):
                s = "in exec"
                # exec_str = get_exec_string(frame.f_back)
                # if exec_str != None:
                #     filename = exec_str
                #     add_quotes_around_file = False
                #     pass
                # pass
            elif not is_pseudo_file:
                s = "in file"
                pass
            pass
        elif not is_pseudo_file:
            s += " called from file"
            pass

        if add_quotes_around_file:
            filename = "'%s'" % filename
        s += " %s at line %s" % (
            format_token(Mformat.Filename, filename, highlight=color),
            format_token(Mformat.LineNumber, "%r" % lineno, highlight=color),
        )
    return s


def frame2file(core_obj, frame, canonic=True):
    if canonic:
        return core_obj.filename(core_obj.canonic_filename(frame))
    else:
        return core_obj.filename(frame.f_code.co_filename)


def frame2filesize(frame):
    if "__cached__" in frame.f_globals:
        bc_path = frame.f_globals["__cached__"]
    else:
        bc_path = None
    path = frame.f_globals["__file__"]
    source_path = getsourcefile(path)
    fs_size = os.stat(source_path).st_size
    if bc_path:
        (
            version,
            timestamp,
            magic_int,
            co,
            is_pypy,
            bc_source_size,
            sip_hash,
        ) = xdis.load_module(bc_path, fast_load=True, get_code=False)
        return fs_size, bc_source_size
    elif osp.exists(path):
        return fs_size, None
    else:
        return None, None


def check_path_with_frame(frame, path):
    my_size = os.stat(path).st_size
    fs_size, bc_size = frame2filesize(frame)
    if bc_size and bc_size != my_size:
        return False, "bytecode and local files mismatch"
    if fs_size and fs_size != my_size:
        return (
            False,
            (
                "frame file size, %d bytes, and local file size, %d bytes, on file %s mismatch"
                % (fs_size, my_size, path)
            ),
        )
    return True, None


def is_exec_stmt(frame):
    """Return True if we are looking at an exec statement"""
    return hasattr(frame, "f_back") and get_call_function_name(frame) == "exec"


import dis

opc = get_opcode(PYTHON_VERSION, IS_PYPY)


def get_call_function_name(frame):
    """If f_back is looking at a call function, return
    the name for it. Otherwise return None"""
    f_back = frame.f_back
    if not f_back:
        return None
    if "CALL_FUNCTION" != Mbytecode.op_at_frame(f_back):
        return None

    co = f_back.f_code
    code = co.co_code
    # labels     = dis.findlabels(code)
    linestarts = dict(dis.findlinestarts(co))
    offset = f_back.f_lasti
    while offset >= 0:
        if offset in linestarts:
            op = code[offset]
            offset += 1
            arg = code[offset]
            # FIXME: put this code in xdis
            extended_arg = 0
            while True:
                if PYTHON_VERSION >= 3.6:
                    if op == opc.EXTENDED_ARG:
                        extended_arg += arg << 8
                        continue
                    arg = code[offset] + extended_arg
                    # FIXME: Python 3.6.0a1 is 2, for 3.6.a3 we have 1
                else:
                    if op == opc.EXTENDED_ARG:
                        extended_arg += arg << 256
                        continue
                    arg = code[offset] + code[offset + 1] * 256 + extended_arg
                break

            if arg < len(co.co_names):
                return co.co_names[arg]
            else:
                return None
        offset -= 1
        pass
    return None


def print_stack_entry(proc_obj, i_stack, color="plain", opts={}):
    frame_lineno = proc_obj.stack[len(proc_obj.stack) - i_stack - 1]
    frame, lineno = frame_lineno
    intf = proc_obj.intf[-1]
    if frame is proc_obj.curframe:
        intf.msg_nocr(format_token(Mformat.Arrow, "->", highlight=color))
    else:
        intf.msg_nocr("##")
    intf.msg(
        "%d %s"
        % (i_stack, format_stack_entry(proc_obj.debugger, frame_lineno, color=color))
    )
    if opts.get("source", False):
        filename = frame2file(proc_obj.core, frame)
        line = linecache.getline(filename, lineno, frame.f_globals)
        intf.msg(line)
        pass

    if opts.get("deparse", False):
        name = frame.f_code.co_name
        if frame.f_lasti == -1:
            last_i = 0
        else:
            last_i = frame.f_lasti
        deparsed, nodeInfo = deparse_offset(frame.f_code, name, last_i, None)
        if name == "<module>":
            name == "module"
        if nodeInfo:
            extractInfo = deparsed.extract_node_info(nodeInfo)
            intf.msg(extractInfo.selectedLine)
            intf.msg(extractInfo.markerLine)
        pass
    if opts.get("full", False):
        names = list(frame.f_locals.keys())
        for name in sorted(names):
            if _with_local_varname.match(name):
                val = frame.f_locals[name]
            else:
                val = proc_obj.getval(name, frame.f_locals)
                pass
            width = opts.get("width", 80)
            Mpp.pp(val, width, intf.msg_nocr, intf.msg, prefix="%s =" % name)
            pass

        deparsed, nodeInfo = deparse_offset(frame.f_code, name, frame.f_lasti, None)
        if name == "<module>":
            name == "module"
        if nodeInfo:
            extractInfo = deparsed.extract_node_info(nodeInfo)
            intf.msg(extractInfo.selectedLine)
            intf.msg(extractInfo.markerLine)
        pass


def print_stack_trace(proc_obj, count=None, color="plain", opts={}):
    "Print count entries of the stack trace"
    if count is None:
        n = len(proc_obj.stack)
    else:
        n = min(len(proc_obj.stack), count)
    try:
        for i in range(n):
            print_stack_entry(proc_obj, i, color=color, opts=opts)
    except KeyboardInterrupt:
        pass
    return


def print_dict(s, obj, title):
    if hasattr(obj, "__dict__"):
        d = obj.__dict__
        if isinstance(d, dict):
            keys = list(d.keys())
            if len(keys) == 0:
                s += "\n  No %s" % title
            else:
                s += "\n  %s:\n" % title
            keys.sort()
            for key in keys:
                s += "    '%s':\t%s\n" % (key, d[key])
                pass
            pass
        pass
    return s


def eval_print_obj(arg, frame, format=None, short=False):
    """Return a string representation of an object """
    try:
        if not frame:
            # ?? Should we have set up a dummy globals
            # to have persistence?
            val = eval(arg, None, None)
        else:
            val = eval(arg, frame.f_globals, frame.f_locals)
            pass
    except:
        return 'No symbol "' + arg + '" in current context.'

    return print_obj(arg, val, format, short)


def print_obj(arg, val, format=None, short=False):
    """Return a string representation of an object """
    what = arg
    if format:
        what = format + " " + arg
        val = Mprint.printf(val, format)
        pass
    s = "%s = %s" % (what, val)
    if not short:
        s += "\n  type = %s" % type(val)
        # Try to list the members of a class.
        # Not sure if this is correct or the
        # best way to do.
        s = print_dict(s, val, "object variables")
        if hasattr(val, "__class__"):
            s = print_dict(s, val.__class__, "class variables")
            pass
        pass
    return s


# Demo stuff above
if __name__ == "__main__":

    class MockDebuggerCore:
        def canonic_filename(self, frame):
            return frame.f_code.co_filename

        def filename(self, name):
            return name

        pass

    class MockDebugger:
        def __init__(self):
            self.core = MockDebuggerCore()
            self.settings = {"maxargstrsize": 80}
            pass

        pass

    frame = inspect.currentframe()
    print(frame2filesize(frame))
    pyc_file = osp.join(
        osp.dirname(__file__), "__pycache__", osp.basename(__file__)[:-3] + ".pyc"
    )
    print(pyc_file, getsourcefile(pyc_file))

    # m = MockDebugger()
    # print(format_stack_entry(m, (frame, 10,)))
    # print(format_stack_entry(m, (frame, 10,), color="dark"))
    # print("frame count: %d" % count_frames(frame))
    # print("frame count: %d" % count_frames(frame.f_back))
    # print("frame count: %d" % count_frames(frame, 1))
    # print("def statement: x=5?: %s" % repr(Mbytecode.is_def_stmt("x=5", frame)))
    # # Not a "def" statement because frame is wrong spot
    # print(Mbytecode.is_def_stmt("def foo():", frame))

    # def sqr(x):
    #     x * x

    # def fn(x):
    #     frame = inspect.currentframe()
    #     print(get_call_function_name(frame))
    #     return

    # print("=" * 30)
    # eval("fn(5)")
    # print("=" * 30)
    # print(print_obj("fn", fn))
    # print("=" * 30)
    # print(print_obj("len", len))
    # print("=" * 30)
    # print(print_obj("MockDebugger", MockDebugger))
    pass



# -*- coding: utf-8 -*-
#   Copyright (C) 2008-2009, 2013, 2015 Rocky Bernstein <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Our local modules
from trepan.processor.command import base_subcmd as Mbase_subcmd


class InfoSource(Mbase_subcmd.DebuggerSubcommand):
    """Information about the current Python file."""

    min_abbrev = 1
    need_stack = True
    short_help = "Information about the current Python file"

    def run(self, args):
        if not self.proc.curframe:
            self.errmsg("No current source file.")
            return
        filename = self.core.canonic_filename(self.proc.curframe)
        self.msg('Current Python file is %s' % self.core.filename(filename))
        return False
    pass



# -*- coding: utf-8 -*-
#   Copyright (C) 2009, 2012-2013, 2015, 2020 Rocky Bernstein
#
# Our local modules
from trepan.processor.command.base_subcmd import DebuggerSubcommand
from trepan.processor.cmdfns import run_set_int


class SetWidth(DebuggerSubcommand):
    """**set width** *number*

Set the number of characters the debugger thinks are in a line.

See also:
--------

`show width`
"""

    in_list = True
    min_abbrev = len("wid")
    short_help = "Set the width of the terminal"

    def run(self, args):
        run_set_int(
            self, " ".join(args), "The 'width' command requires a line width", 0, None
        )
        return

    pass


if __name__ == "__main__":
    from trepan.processor.command.set_subcmd.__demo_helper__ import demo_run

    sub = demo_run(SetWidth, ["10"])
    d = sub.proc.debugger
    print(d.settings["width"])
    sub.run(["100"])
    print(d.settings["width"])
    pass



# -*- coding: utf-8 -*-
#  Copyright (C) 2009, 2013, 2015, 2020 Rocky Bernstein
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Our local modules
from trepan.processor.command.base_cmd import DebuggerCommand


class HandleCommand(DebuggerCommand):
    """**handle** [*signal-name* [*action1* *action2* ...]]

Specify how to handle a signal *signal-name*. *signal-name* can be a
signal name like `SIGINT` or a signal number like 2. The absolute
value is used for numbers so -9 is the same as 9 (`SIGKILL`). When
signal names are used, you can drop off the leading "SIG" if you want. Also
letter case is not important either.

Arguments are signals and actions to apply to those signals.
recognized actions include `stop`, `nostop`, `print`, `noprint`,
`pass`, `nopass`, `ignore`, or `noignore`.

`stop` means reenter debugger if this signal happens (implies `print` and
`nopass`).

`Print` means print a message if this signal happens.

`Pass` means let program see this signal; otherwise the program see it.

`Ignore` is a synonym for `nopass`; `noignore` is a synonym for `pass`.

Without any action names the current settings are shown.

**Examples:**

  handle INT         # Show current settings of SIGINT
  handle SIGINT      # same as above
  handle int         # same as above
  handle 2           # Probably the same as above
  handle -2          # the same as above
  handle INT nostop  # Don't stop in the debugger on SIGINT
"""

    short_help = "Specify how to handle a signal"

    DebuggerCommand.setup(locals(), category="running", min_args=1)

    def run(self, args):
        if self.debugger.sigmgr.action(" ".join(args[1:])) and len(args) > 2:
            # Show results of recent change
            self.debugger.sigmgr.info_signal([args[1]])
            pass
        return

    pass


if __name__ == "__main__":
    from trepan import debugger as Mdebugger

    d = Mdebugger.Trepan()
    command = HandleCommand(d.core.processor)
    command.run(["handle", "USR1"])
    command.run(["handle", "term", "stop"])
    pass



# -*- coding: utf-8 -*-
#   Copyright (C) 2009, 2013, 2020 Rocky Bernstein
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

from trepan.processor.command.base_submgr import SubcommandMgr


class ShowCommand(SubcommandMgr):
    """**show** *subcommand*

Generic command for showing things about the debugger.  You can
give unique prefix of the name of a subcommand to get information
about just that subcommand.

Type `show` for a list of *show* subcommands and what they do.
Type `help show *` for just a list of *show* subcommands.
"""

    short_help = "Show parts of the debugger environment"
    SubcommandMgr.setup(locals(), category="status")


if __name__ == "__main__":
    from trepan.processor.command import mock

    d, cp = mock.dbg_setup()
    command = ShowCommand(cp, "show")
    command.run(["show"])
    pass



# -*- coding: utf-8 -*-
#   Copyright (C) 2009, 2013-2015 Rocky Bernstein <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Debugger FIFO Input/Output interface. """

import os

if hasattr(os, 'mkfifo'):
    import atexit, tempfile

    from trepan import misc as Mmisc
    from trepan.inout.base import DebuggerInOutBase

    # FIXME: Consider using Python's socketserver/SocketServer?
    class FIFOServer(DebuggerInOutBase):
        """Debugger Server Input/Output Socket."""

        DEFAULT_INIT_OPTS = {'open': True}

        def __init__(self, opts=None):
            get_option = lambda key: Mmisc.option_set(opts, key,
                                                      self.DEFAULT_INIT_OPTS)
            atexit.register(self.close)
            self.flush_after_write = True
            self.line_edit = False  # Our name for GNU readline capability
            self.in_name   = None   # String: input file name
            self.input     = None   # File Descriptor
            self.out_name  = None   # String: output file name
            self.output    = None   # String: output file name
            self.state     = 'disconnected'
            if get_option('open'):
                self.open(opts)
                pass
            return

        def close(self):
            """ Closes both input and output. """
            self.state = 'closing'
            if self.input:
                self.input.close()
                pass
            if self.in_name and os.path.exists(self.in_name):
                os.unlink(self.in_name)
                pass
            if self.output:
                self.output.close()
                pass
            if self.out_name and os.path.exists(self.out_name):
                os.unlink(self.out_name)
                pass
            self.state = 'disconnnected'
            return

        def flush(self):
            return self.output.flush()

        def open(self, opts=None):
            d              = tempfile.gettempdir()
            pid            = os.getpid()
            self.out_name  = os.path.join(d, ('trepan-%s.out' % pid))
            self.in_name   = os.path.join(d, ('trepan-%s.in' % pid))
            try:
                os.mkfifo(self.in_name)
                os.mkfifo(self.out_name)
                self.state     = 'active'
            except OSError:
                self.state = 'error'
                pass
            return

        def read_msg(self):
            """Read a line of input. EOFError will be raised on EOF.

            Note that we don't support prompting"""
            # FIXME: do we have to create and check a buffer for
            # lines?
            if self.state == 'active':
                if not self.input:
                    self.input = open(self.in_name, 'r')
                    pass
                line = self.input.readline()
                if not line:
                    self.state = 'disconnected'
                    raise EOFError
                return line.rstrip("\n")
            else:
                raise EOFError
            return  # Not reached

        def write(self, msg):
            """ This method the debugger uses to write. In contrast to
            writeline, no newline is added to the end to `str'.
            """
            if self.state == 'active':
                if not self.output:
                    self.output = open(self.out_name, 'w')
                    pass
                pass
            else:
                raise EOFError
            self.output.write(msg)
            if self.flush_after_write: self.flush()
            return
    # Demo
    if __name__=='__main__':
        fifo = FIFOServer(opts={'open': False})
        import sys
        if len(sys.argv) > 1:
            fifo.open()
            print('Looking for input on %s"...' % fifo.in_name)
            while True:
                try:
                    fifo.write('nu?')
                    fifo.writeline(' ')
                    line = fifo.readline()
                    print(line)
                    fifo.writeline('ack: ' + line)
                except EOFError:
                    break
                pass
            pass
        fifo.close()
        pass
    pass



#  Copyright (C) 2013, 2015, 2020 Rocky Bernstein
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Our local modules
from trepan.processor.command.base_cmd import DebuggerCommand


class AliasCommand(DebuggerCommand):
    """**alias** *alias-name* *debugger-command*

Add alias *alias-name* for a debugger command *debugger-command*.  You
might do this if you want shorter command names or more commands that
have more familiar meanings.

Another related use is as a command abbreviation for a command that
would otherwise be ambiguous. For example, by default we make `s` be
an alias of `step` to force it to be used. Without the alias, `s`
might be `step`, `show`, or `set` among others.

Examples:
--------

    alias cat list   # "cat prog.py" is the same as "list prog.py"
    alias s   step   # "s" is now an alias for "step".
                     # The above example is done by default.

See also:
---------

`unalias` and `show alias`.
    """

    name = "alias"
    short_help = "Add an alias for a debugger command"

    DebuggerCommand.setup(locals(), category="support", max_args=2, need_stack=True)

    # Run command.
    def run(self, args):
        if len(args) == 1:
            self.proc.commands["show"].run(["show", "alias"])
        elif len(args) == 2:
            self.proc.commands["show"].run(["show", "alias", args[1]])
        else:
            junk, al, command = args
            if command in self.proc.commands:
                if al in self.proc.aliases:
                    old_command = self.proc.aliases[al]
                    self.msg(
                        (
                            "Alias '%s#' for command '%s'replaced old "
                            + "alias for '%s'."
                        )
                        % (al, command, old_command)
                    )
                else:
                    self.msg("New alias '%s' for command '%s' created." % (al, command))
                    pass
                self.proc.aliases[al] = command
            else:
                self.errmsg(
                    ("You must alias to a command name, and '%s' " + "and is not one.")
                    % command
                )
                pass
            return
        pass

    pass


if __name__ == "__main__":
    # Demo it.
    from trepan.processor.command import mock

    d, cp = mock.dbg_setup()
    command = AliasCommand(cp)
    command.run(["alias", "yy", "foo"])
    command.run(["alias", "yy", " foo"])
    command.run(["alias", "yy" "step"])
    command.run(["alias"])
    command.run(["alias", "yy" "next"])
    pass



# -*- coding: utf-8 -*-
#
#  Copyright (C) 2009-2010, 2013, 2015-2018, 2020 Rocky Bernstein
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

import inspect
from pyficache import code_line_info, code_offset_info
from trepan.misc import wrapped_lines, pretty_modfunc_name
from trepan.processor.parse.semantics import build_bp_expr
from trepan.processor.parse.parser import LocationError
from trepan.processor.parse.scanner import ScannerError
from trepan.processor.location import resolve_location


def set_break(
    cmd_obj,
    func,
    filename,
    lineno,
    condition,
    temporary,
    args,
    force=False,
    offset=None,
):
    if lineno is None and offset is None:
        part1 = "I don't understand '%s' as a line number, offset, or function name," % " ".join(
            args[1:]
        )
        msg = wrapped_lines(
            part1, "or file/module plus line number.", cmd_obj.settings["width"]
        )
        cmd_obj.errmsg(msg)
        return False
    if filename is None:
        filename = cmd_obj.proc.curframe.f_code.co_filename
        filename = cmd_obj.core.canonic(filename)
        pass
    if func is None:
        if lineno:
            line_info = code_line_info(filename, lineno)
            if not line_info:
                part1 = "File %s" % cmd_obj.core.filename(filename)
                msg = wrapped_lines(
                    part1,
                    "is not stoppable at line %d." % lineno,
                    cmd_obj.settings["width"],
                )
                cmd_obj.errmsg(msg)
                if force:
                    cmd_obj.msg("Breakpoint set although it may never be reached")
                else:
                    return False
        else:
            assert offset is not None
            lineno = code_offset_info(filename, offset)
            if lineno is None:
                part1 = "File %s" % cmd_obj.core.filename(filename)
                msg = wrapped_lines(
                    part1,
                    "has no line associated with offset %d." % offset,
                    cmd_obj.settings["width"],
                )
                cmd_obj.errmsg(msg)
                return False

        pass
    bp = cmd_obj.core.bpmgr.add_breakpoint(
        filename,
        lineno=lineno,
        offset=offset,
        temporary=temporary,
        condition=condition,
        func=func)
    if func and inspect.isfunction(func):
        cmd_obj.msg(
            "Breakpoint %d set on calling function %s()" % (bp.number, func.__name__)
        )
        part1 = "Currently this is line %d of file" % lineno
        msg = wrapped_lines(
            part1, cmd_obj.core.filename(filename), cmd_obj.settings["width"]
        )
        cmd_obj.msg(msg)
    else:
        part1 = ("Breakpoint %d set at line %d of file" %
                 (bp.number, lineno))
        msg = wrapped_lines(
            part1, cmd_obj.core.filename(filename), cmd_obj.settings["width"]
        )
        cmd_obj.msg(msg)
        if func:
            func_str = " of %s" % pretty_modfunc_name(func)
        else:
            func_str = ""
        if offset is not None and offset >= 0:
            cmd_obj.msg("Breakpoint is at offset %d%s "% (offset, func_str))
        pass
    return True


INVALID_PARSE_BREAK = (None, None, None, None, None)


def parse_break_cmd(proc, args):
    if proc.current_command is None:
        proc.errmsg("Internal error")
        return INVALID_PARSE_BREAK

    text = proc.current_command[len(args[0]) + 1 :]
    if len(args) > 1 and args[1] == "if":
        location = "."
        condition = text[text.find("if ") + 3 :]
    elif text == "":
        location = "."
        condition = None
    else:
        try:
            bp_expr = build_bp_expr(text)
        except LocationError as e:
            proc.errmsg("Error in parsing breakpoint expression at or around:")
            proc.errmsg(e.text)
            proc.errmsg(e.text_cursor)
            return INVALID_PARSE_BREAK
        except ScannerError as e:
            proc.errmsg("Lexical error in parsing breakpoint expression at or around:")
            proc.errmsg(e.text)
            proc.errmsg(e.text_cursor)
            return INVALID_PARSE_BREAK

        location = bp_expr.location
        condition = bp_expr.condition

    location = resolve_location(proc, location)
    if location:
        return (
            location.method,
            location.path,
            location.line_number,
            condition,
            location.offset,
        )
    else:
        return INVALID_PARSE_BREAK


# Demo it
if __name__ == "__main__":
    from trepan.processor.command.mock import MockDebugger
    from trepan.processor.cmdproc import CommandProcessor
    import sys

    d = MockDebugger()
    cmdproc = CommandProcessor(d.core)
    # print '-' * 10
    # print_source_line(sys.stdout.write, 100, 'source_line_test.py')
    # print '-' * 10
    cmdproc.frame = sys._getframe()
    cmdproc.setup()
    for cmd in (
        "break '''c:\\tmp\\foo.bat''':1",
        'break """/Users/My Documents/foo.py""":2',
        "break",
        "break 10",
        "break if True",
        "break cmdproc.py:5",
        "break set_break()",
        "break 4 if i==5",
        # "break cmdproc.setup()",
    ):
        args = cmd.split(" ")
        cmdproc.current_command = cmd
        print(parse_break_cmd(cmdproc, args))
    pass



# -*- coding: utf-8 -*-
#  Copyright (C) 2010, 2012-2015, 2020 Rocky Bernstein
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys, threading
from trepan.processor.command.base_cmd import DebuggerCommand
from trepan.processor.complete import complete_identifier


class DebugCommand(DebuggerCommand):
    """**debug** *python-expression*

Enter a nested debugger that steps through the *python-expression* argument
which is an arbitrary expression to be executed the current
environment."""

    short_help = "Debug *python-expression"

    complete = complete_identifier

    DebuggerCommand.setup(locals(), category="support", min_args=1, need_stack=True)

    def run(self, args):
        arg = " ".join(args[1:])
        curframe = self.proc.curframe
        if not curframe:
            self.msg("No frame selected.")
            return

        for attr in ("prompt_str", "frame", "event", "event_arg", "curindex"):
            cmd = "old_%s = self.proc.%s" % (attr, attr)
            exec(cmd)
            if hasattr(self.proc, "print_location"):
                self.proc.print_location()
            pass

        old_lock = self.core.debugger_lock
        old_stop_level = self.core.stop_level
        old_different_line = self.core.stop_level
        self.proc.debug_nest += 1

        self.core.debugger_lock = threading.Lock()
        self.core.stop_level = None
        self.core.different_line = None
        global_vars = curframe.f_globals
        local_vars = curframe.f_locals

        self.section("ENTERING NESTED DEBUGGER")

        self.core.step_ignore = 2  # call_tracing will stop in itself.
        try:
            ret = sys.call_tracing(eval, (arg, global_vars, local_vars))
            self.msg("R=> %s" % self.proc._saferepr(ret))
        except:
            pass
        self.section("LEAVING NESTED DEBUGGER")

        self.core.debugger_lock = old_lock
        self.core.stop_level = old_stop_level
        self.core.different_line = old_different_line
        self.proc.continue_running = False
        self.proc.debug_nest -= 1

        for attr in ("prompt_str", "frame", "event", "event_arg", "curindex"):
            cmd = "self.proc.%s = old_%s" % (attr, attr)
            exec(cmd)
            pass
        self.proc.location()
        return False

    pass


if __name__ == "__main__":
    import inspect
    from trepan.processor import cmdproc as Mcmdproc
    from trepan import debugger

    d = debugger.Trepan()
    cp = d.core.processor
    cp.curframe = inspect.currentframe()
    cp.stack, cp.curindex = Mcmdproc.get_stack(cp.curframe, None, None, cp)
    command = DebugCommand(cp)

    def test_fn():
        return 5

    command.run(["debug", "test_fn()"])
    pass



# -*- coding: utf-8 -*-
#  Copyright (C) 2009, 2015, 2020 Rocky Bernstein
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Our local modules
from trepan.processor.command.base_subcmd import DebuggerShowBoolSubcommand


class ShowAutoList(DebuggerShowBoolSubcommand):
    """**show autolist**

Show debugger `list` command automatically on entry.

See also:
---------

`set autolist`"""

    min_abbrev = len("autol")
    short_help = "Show `list` on debugger entry"
    pass


if __name__ == "__main__":
    from trepan.processor.command.show_subcmd.__demo_helper__ import demo_run

    demo_run(ShowAutoList)
    pass



#!/usr/bin/env python3
'Unit test for trepan.lib.printing'
import unittest

from trepan.lib import printing as Mprint


class TestLibPrint(unittest.TestCase):

    def test_lib_printf(self):
        self.assertEqual('0o37', Mprint.printf(31, "/o"))
        self.assertEqual('00011111', Mprint.printf(31, "/t"))
        self.assertEqual('!', Mprint.printf(33, "/c"))
        self.assertEqual('0x21', Mprint.printf(33, "/x"))
        return

    def test_lib_argspec(self):
        self.assertEqual('test_lib_argspec(self)',
                         Mprint.print_argspec(self.test_lib_argspec,
                                              'test_lib_argspec'))
        self.assertFalse(Mprint.print_argspec(None, 'invalid_fn'))
        return
    pass

if __name__ == '__main__':
    unittest.main()
